"""Parameterized model abstraction."""


class ParameterizedModel(object):
    """A parameterized model.

    Parameters
    ----------
    model_getter : callable
        A function that returns an estimator object implementing ‘fit’ when
        supplied with values for its required keyword arguments, ignoring any
        unkown keyword arguments.
    param_grid : iterable over dict of string to any
        An iterable over parameter realizations. The ParameterGrid sklearn
        class is a great example.
    model_id_getter : callable, optional
        A function that returns an id string when supplied with values for its
        required keyword arguments, ignoring any unkown keyword arguments.
    """

    @staticmethod
    def _default_model_id_getter(**kwargs):
        param_strings = ['{}={}'.format(key, kwargs[key]) for key in kwargs]
        return '_'.join(param_strings)

    def __init__(self, model_getter, param_grid, model_id_getter=None):
        if model_id_getter is None:
            model_id_getter = ParameterizedModel._default_model_id_getter
        self._model_getter = model_getter
        self._param_grid = param_grid
        self._model_id_getter = model_id_getter

    def __iter__(self):
        """Iterate on all realized models induced by this parameterized model.

        Returns
        -------
        models : iterator over sklearn.BaseEstimator objects
            Yields scikit-learn estimator objects.
        """
        for params in self._param_grid:
            yield self._model_getter(**params)

    def model_n_params_iter(self):
        """Iterate on all model and parameters pairs induced by this
        parameterized model.

        Returns
        -------
        models : iterator over tuples
            Yields 2-tuples of an estimator object and a dict.
        """
        for params in self._param_grid:
            yield self._model_getter(**params), params

    def model_by_params(self, params):
        """Returns a realized model by the given params.

        Parameters
        ----------
        params : dict of string to any
            Parameter assignments by which to produce a model.

        Returns
        -------
        sklearn.BaseEstimator
            An estimator.
        """
        return self._model_getter(**params)

    def model_id_by_params(self, params):
        """Returns a model id by the given params.

        Parameters
        ----------
        params : dict of string to any
            Parameter assignments by which to produce a model.

        Returns
        -------
        string
            A model id string.
        """
        return self._model_id_getter(**params)


class ConstrainedParameterizedModel(ParameterizedModel):
    """A constrained parameterized model.

    Parameters
    ----------
    model_getter : function
        A function that returns an estimator object implementing ‘fit’ when
        supplied with values for its required keyword arguments, ignoring any
        unkown keyword arguments.
    param_grid : skutil.model_selection.ConstrainedParameterGrid
        A constrained grid over model parameters.
    """

    def partial(self, assign_grid):
        """Returns a new parameterized model by the given partial assignment.

        Parameters
        ----------
        assign_grid : dict of string to object or sequence
            A, possibly partial, assignment to the parameters of the parameter
            grid. Keys not appearing in this grid are ignored.

        Returns
        -------
        ConstrainedParameterizedModel
            A new constrained parameterized model induced by the given partial
            assignment.
        """
        return ConstrainedParameterizedModel(
            model_getter=self._model_getter,
            param_grid=self._param_grid.partial(assign_grid),
            model_id_getter=self._model_id_getter,
        )



"""Shared testing resources for folk."""

import functools

import pandas as pd
from sklearn.linear_model import LogisticRegression

import pdpipe as pdp
from birch import Birch
from pymongo import MongoClient
from skutil.model_selection import ConstrainedParameterGrid

from folk import (
    ConstrainedParameterizedModel,
    ParameterizedPipeline,
)
from folk.cfg import CfgKey


# === Constants ===

def _model_getter(penalty, C, **kwargs):
    return LogisticRegression(penalty=penalty, C=C)


MODEL_PGRID = ConstrainedParameterGrid({
    'penalty': ['l1', 'l2'],
    'C': [0.3, 0.6, 0.9],
})


PMODEL = ConstrainedParameterizedModel(
    model_getter=_model_getter,
    param_grid=MODEL_PGRID,
)


def _pipeline_getter(lower=False, **kwargs):
    stages = [
        pdp.ColRename({'shleem_count': 'scount'}),
    ]
    if lower:
        stages.append(pdp.ApplyByCols('rank', str.lower))
    stages.append(pdp.ColDrop(columns='id'))
    return pdp.Pipeline(stages)


PIPE_PGRID = ConstrainedParameterGrid({
    'lower': [True, False],
    'lbl_col': ['rank'],
})


PPIPELINE = ParameterizedPipeline(
    pipeline_getter=_pipeline_getter,
    param_grid=PIPE_PGRID,
)

TEST_METRICS_DB = 'FOLK_TEST'


def _test_df():
    return pd.DataFrame(
        data=[
            [23, 'A', 4.3, 'e2f'],
            [43, 'C', 3.5, '2eqef'],
            [80, 'B', 8.8, 'isudgv'],
            [13, 'A', 3.6, '239u8h'],
            [77, 'B', 5.0, 'ds9h77'],
            [103, 'B', 5.5, '9hudsg'],
            [33, 'C', 2.2, '3t782'],
            [42, 'C', 3.66, 'dh2u'],
        ],
        index=[0, 1, 2, 3, 4, 5, 6, 7],
        columns=['shleem_count', 'rank', 'hizzard_ratio', 'id'],
    )


@functools.lru_cache(maxsize=1)
def _metrics_collection():
    folk_cfg = Birch('folk')
    db_cfg = folk_cfg[CfgKey.MetricDbs][TEST_METRICS_DB]
    client = MongoClient(db_cfg['URI'])
    metrics_collection = client[db_cfg['DB_NAME']][db_cfg['COLLECTION_NAME']]
    return metrics_collection


def _clean_db():
    metrics_collection = _metrics_collection()
    metrics_collection.delete_many({})
    assert metrics_collection.count({}) == 0



"""Folk configuration."""

import functools

from birch import Birch


@functools.lru_cache(maxsize=2)
def folk_cfg():
    return Birch(
        namespace='folk',
        supported_formats=['json', 'yaml'],
    )


class CfgKey(object):
    MetricDbNames = "METRIC_DB_NAMES"
    MetricDbs = "METRIC_DBS"
    DbType = "TYPE"



"""Test folk's evaluate module."""

import pytest


from folk import (
    eval_param_pipeline_n_model,
)

from .shared import (
    MODEL_PGRID,
    PIPE_PGRID,
    PPIPELINE,
    PMODEL,
    TEST_METRICS_DB,
    _test_df,
    _metrics_collection,
    _clean_db,
)


@pytest.fixture(scope="session", autouse=True)
def prep_and_teardown(request):
    # Will be executed before the first test
    _clean_db()

    yield

    # Will be executed after the last test
    _clean_db()


def test_base_eval():
    eval_param_pipeline_n_model(
        param_pipeline=PPIPELINE,
        param_model=PMODEL,
        dataset=_test_df(),
        metric_db=TEST_METRICS_DB,
        n_folds=2,
    )


if __name__ == "__main__":
    test_base_eval()
    model_permutations = len(MODEL_PGRID)
    pipeline_permutations = len(PIPE_PGRID)
    total_experiments = pipeline_permutations * model_permutations
    metric_collection = _metrics_collection()
    assert metric_collection.count({}) == total_experiments



"""Parameterized pipeline abstraction."""


class ParameterizedPipeline(object):
    """A parameterized pipeline.

    This class was written with pdpipe.Pipeline objects in mind, but can be
    used with all callables that transform pandas.DataFrame objects.

    Parameters
    ----------
    pipeline_getter : function
        A function that returns a pipeline for pandas DataFrame objects when
        supplied with values for its required keyword arguments, ignoring any
        unkown keyword arguments.
    param_grid : iterable over dict of string to any
        An iterable over parameter realizations. The ParameterGrid sklearn
        class is a great example.
    """

    def __init__(self, pipeline_getter, param_grid):
        self.pipeline_getter = pipeline_getter
        self.param_grid = param_grid

    def __iter__(self):
        """Iterate over all pipelines induced by this parameterized pipeline.

        Returns
        -------
        models : iterator over callable objects
            Yields callable objects that can be applied to dataframe objects
            to produce new dataframes.
        """
        for params in self.param_grid:
            yield self.pipeline_getter(**params)

    def pipe_n_params_iter(self):
        """Iterate on all pipeline and parameters pairs induced by this
        parameterized pipieline.

        Returns
        -------
        models : iterator over tuples
            Yields 2-tuples of a pdpipe.Pipeline and a dict.
        """
        for params in self.param_grid:
            yield self.pipeline_getter(**params), params

    def pipeline_by_params(self, params):
        """Returns a realized pipeline by the given params.

        Parameters
        ----------
        params : dict of string to any
            Parameter assignments by which to produce a pipeline.

        Returns
        -------
        callable
            A callable object that can be applied to dataframe objects
            to produce new dataframes.
        """
        return self.pipeline_getter(**params)



"""Test model-related folk stuff."""

from .shared import (
    LogisticRegression,
    PMODEL,
)


def test_model_base():
    for model in PMODEL:
        assert isinstance(model, LogisticRegression)
    model = PMODEL.model_by_params({'penalty': 'l1', 'C': 0.3})
    assert isinstance(model, LogisticRegression)



"""Utilities for pandas."""

from .model import (  # noqa: F401
    ParameterizedModel,
    ConstrainedParameterizedModel,
)
from .pipe import ParameterizedPipeline  # noqa: F401
from .metricsdb import MetricKey  # noqa: F401
from .evaluate import (  # noqa: F401
    eval_param_pipeline_n_model,
)

from ._version import get_versions
__version__ = get_versions()['version']

for name in ['get_versions', '_version', 'model', 'pipe', 'name']:
    try:
        globals().pop(name)
    except KeyError:
        pass



"""Metric databases for folk."""

import abc
import warnings
from datetime import datetime
from importlib import import_module

from decore import lazy_property
from strct.dicts import safe_nested_val

from .cfg import (
    folk_cfg,
    CfgKey,
)
from .exceptions import (
    FolkMissingConfigurationValueError,
)


class MetricKey(object):
    # folk-specific parameters
    MODEL_ID = 'model_identifier'
    RUN_ID = 'run_id'
    RUN_AT = 'run_at'
    # general ML parameters
    LBL_COL = 'lbl_col'
    DATASET_SIZE = 'dataset_size'
    N_FOLDS = 'n_folds'
    N_CLASS = 'n_classes'
    ACC_MEAN = 'accuracy_mean'
    ACC_STD = 'accuracy_std'


class FolkMetricsDB(object, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def write_experiment_res(self, run_id=None, **kwargs):
        """Writes the result of a single experiment to the database.

        Parameters
        ----------
        res_doc : dict
            A key-value mapping of parameters and their values.
        run_id : str, optional
            A string identifier for the run this experiment is part of.
        """
        pass  # pragma: no cover


class FolkMetricsMongoDB(FolkMetricsDB):

    _ERR = "Missing config value for {} for MongoDB-based {} folk metrics db."

    """A MongoDB-based folk metrics database.

    Parameters
    ----------
    db_cfg : dict
        A mapping of db configuration parameters to their values.
    """
    def __init__(self, name, db_cfg):
        self.name = name
        for cfg_param in ['URI', 'DB_NAME', 'COLLECTION_NAME']:
            try:
                setattr(self, cfg_param.lower(), db_cfg[cfg_param])
            except KeyError:
                raise FolkMissingConfigurationValueError(
                    FolkMetricsMongoDB._ERR.format(cfg_param, self.name))
        self.collection = None

    @staticmethod
    @lazy_property
    def _mongo_client_cls():
        pymongo = import_module('pymongo')
        submodule = getattr(pymongo, 'mongo_client')
        return getattr(submodule, 'MongoClient')

    def get_collection(self):
        if self.collection is None:
            MongoClient = FolkMetricsMongoDB._mongo_client_cls()
            client = MongoClient(self.uri)
            db = client[self.db_name]
            self.collection = db[self.collection_name]
        return self.collection

    def write_experiment_res(self, res_doc, run_id=None):
        run_at = datetime.utcnow()
        if run_id is None:
            run_id = str(run_at.timestamp()).replace('.', '')
        col_obj = self.get_collection()
        col_obj.insert_one({
            MetricKey.RUN_ID: run_id,
            MetricKey.RUN_AT: run_at,
            **res_doc,
        })


TYPE_TO_CLS_MAP = {
    'mongodb': FolkMetricsMongoDB,
}


NAME_TO_DB_MAP = {}


def populate_name_to_db_map():
    cfg = folk_cfg()
    names_str = cfg.get(CfgKey.MetricDbNames, None)
    if names_str is None:
        return
    for name in names_str.split(','):
        db_cfg = safe_nested_val(
            key_tuple=(CfgKey.MetricDbs, name),
            dict_obj=cfg,
            default_value=None,
        )
        if db_cfg is None:
            warnings.warn(
                "Folk: No configuration for db {} included in {}.".format(
                    name, CfgKey.MetricDbs))
            continue
        try:
            db_type = db_cfg[CfgKey.DbType]
        except KeyError:
            warnings.warn("Folk: No type configured for db {}.".format(name))
            continue
        try:
            db_cls = TYPE_TO_CLS_MAP[db_type.lower()]
        except KeyError:
            warnings.warn("Folk: Unkown type configured for db {}.".format(
                name))
            continue
        db_instance = db_cls(name=name, db_cfg=db_cfg)
        NAME_TO_DB_MAP[name] = db_instance


populate_name_to_db_map()


def write_experiment_res(res_doc, db_name, run_id=None):
    """Writes an experiment results document to a folk metrics database.

    Parameters
    ----------
    res_doc : dict
        A key-value mapping of parameters and their values.
    db_name : str
        The name of the folk metrics db to write to. A db with this name must
        be configured for folk.
    run_id : str, optional
        A string identifier for the run this experiment is part of.
    """
    try:
        db_obj = NAME_TO_DB_MAP[db_name]
    except KeyError:
        warnings.warn((
            "Folk: No intact configuration for db {}. "
            "Results were not written to db.").format(db_name))
        return
    db_obj.write_experiment_res(res_doc=res_doc, run_id=run_id)



"""Evalute folk parameterized pipelines and models."""

from datetime import datetime

from pdutil.transform import x_y_by_col_lbl
from sklearn.model_selection import (
    cross_val_score,
)

from .metricsdb import (
    MetricKey,
    write_experiment_res,
)


def eval_model_by_params(
        run_id, model, model_id, df, lbl_col, metric_db=None, n_folds=None,
        n_jobs=None, **kwargs):
    if n_folds is None:
        n_folds = 5
    if n_jobs is None:
        n_jobs = 1
    print("  - Testing {}...".format(model_id))
    X, y = x_y_by_col_lbl(df, lbl_col)
    print('    Performing {}-fold cross validation...'.format(n_folds))
    scores = cross_val_score(
        model, X=X, y=y, cv=n_folds, scoring='accuracy',
        n_jobs=n_jobs,
    )
    acc_mean = scores.mean()
    acc_std = scores.std()
    print("    Accuracy: {:.2f} (+/- {:.2f})".format(
        scores.mean(), scores.std() * 2))
    n_classes = len(df[lbl_col].unique())
    if metric_db:
        print("    Writing results to db...")
        write_experiment_res(
            res_doc={
                MetricKey.MODEL_ID: model_id,
                MetricKey.LBL_COL: lbl_col,
                MetricKey.ACC_MEAN: acc_mean,
                MetricKey.ACC_STD: acc_std,
                MetricKey.N_FOLDS: n_folds,
                MetricKey.DATASET_SIZE: len(df),
                MetricKey.N_CLASS: n_classes,
                **kwargs,
            },
            db_name=metric_db,
            run_id=run_id,
        )


def eval_pmodel_by_params(
        run_id, pipeline, pmodel, raw_df, metric_db=None, n_folds=None,
        n_jobs=None, **kwargs):
    print("=============================")
    print("Testing parameterized model on pipeline with "
          "params {}".format(kwargs))
    df = pipeline.fit_transform(raw_df)
    print("Dataset size: {}".format(len(df)))
    print("Number of columns: {}".format(len(df.columns)))
    print("Resulting dataset size: {}".format(len(df)))
    print("Resulting columns: {}".format(sorted(list(df.columns))))
    partial_pmodel = pmodel.partial(kwargs)
    j = 1
    for model, params in partial_pmodel.model_n_params_iter():
        model_id = pmodel.model_id_by_params(params)
        print("-------- Model {} --------".format(j))
        print("Testing model with params: {}".format(params))
        full_params = {
            'run_id': run_id,
            'model': model,
            'model_id': model_id,
            'metric_db': metric_db,
            'df': df,
            'n_folds': n_folds,
            'n_jobs': n_jobs,
        }
        full_params.update(params)
        full_params.update(kwargs)
        eval_model_by_params(
            **full_params
        )
        j += 1
    print("=============================\n")


def eval_param_pipeline_n_model(param_pipeline, param_model, dataset,
                                metric_db=None, n_folds=None, n_jobs=None):
    """Evaluates the given parameterized pipeline and model.

    Parameters
    ----------
    param_pipeline : folk.ParameterizedPipeline
        The parameterized pipeline to evaluate.
    param_model : folk.ParameterizedModel
        The parameterized model to evaluate.
    dataset : pandas.DataFrame
        The dataset to evaluate the parameterized pipeline and model on.
    metric_db : str, optional
        The name of the folk metrics db to write results to. Must be included
        in folk's configuration. If not given, results are not written to db.
    n_folds
    n_jobs
    """
    run_at = datetime.utcnow()
    run_id = str(run_at.timestamp()).replace('.', '')
    i = 1
    for pipeline, params in param_pipeline.pipe_n_params_iter():
        print("Pipeline #{}".format(i))
        eval_pmodel_by_params(
            run_id=run_id,
            pipeline=pipeline,
            pmodel=param_model,
            raw_df=dataset,
            metric_db=metric_db,
            n_folds=n_folds,
            n_jobs=n_jobs,
            **params,
        )
        i += 1



"""Test pipeline-related folk stuff."""

import pdpipe as pdp

from .shared import PPIPELINE


def test_pipe_base():
    for pipe in PPIPELINE:
        assert isinstance(pipe, pdp.Pipeline)
    pipe = PPIPELINE.pipeline_by_params({'lower': True})
    assert isinstance(pipe, pdp.Pipeline)
